/*
 * jobd.c - JOBs manager daemon
 *
 * Author: Guillaume Thouvenin <guillaume.thouvenin@bull.net>
 */

#include <assert.h>		/* for assertion */
#include <dirent.h>		/* opendir() */
#include <errno.h>		/* errno */
#include <fcntl.h>		/* open(), opendir() */
#include <glib.h>		/* for Glist */
#include <signal.h>		/* sigaction() */
#include <stdio.h>		/* perror() */
#include <stdlib.h>		/* exit() */
#include <string.h>		/* strerror() */
#include <syslog.h>		/* syslog() */
#include <unistd.h>		/* sysconf(), setsid(), fork(), dup(), close() */

#include <asm/types.h>

#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/types.h>		/* fork(), open(), waitpid(), IPC */
#include <sys/un.h>

#include <linux/netlink.h>
/* 
 * As connector is not included in the libc we need to include it through
 * kernel source. The problem is that there is a conflit with some 
 * definitions in time.h. Some variable are defined in sys/time.h and
 * redefined in linux/time.h (included in cn_proc.h). Thus we add this
 * ugly definition to be able to compile. We will remove that hack as 
 * soon as libc includes connectors.
 */
#define _LINUX_TIME_H
#include <linux/cn_proc.h>
#include <linux/connector.h>
#undef _LINUX_TIME_H

#include "jobd.h"

#ifdef DEBUG_JOBD
#define JOBD_DEBUG(args...)	syslog(LOG_INFO, args)
#else
#define JOBD_DEBUG(args...)	do {} while(0)
#endif				/* DEBUG_JOBD */

#define JOBD_LOGIDENT	"jobd"
#define PROC_BASE	"/proc"

#define MESSAGE_SIZE    (sizeof(struct nlmsghdr) + \
                         sizeof(struct cn_msg)   + \
                         sizeof(int))

/* item stored in jobs_ht */
struct ht_elmt {
	pid_t pid;		/* process ID (the key for the hash table ) */
	GList *jobs;		/* list of jobs to which it belongs */
};

/* used to exchange data between the job daemon and a user space application */
struct send_list_data {
	int index;
	int sock;
};

/*
 * structures to store jobs information.
 * jobs_ht allows quick access to process information by its pid while
 * jobs_list allows access to a list of PID for given jobID.
 */
static GHashTable *jobs_ht;
static GList *jobs_list;

static unsigned int next_jid;	/* next_jid = 0 */
static int sk_nl;		/* socket used by netlink */
static int sk_unix;		/* socket used for local communication */

static const char *log_ident = JOBD_LOGIDENT;


/**
 * cn_fork_listen - 
 */
static inline void cn_fork_listen(int sock)
{
	char buff[128];		/* must be > MESSAGE_SIZE */
	struct nlmsghdr *hdr;
	struct cn_msg *msg;

	/* Clear the buffer */
	memset(buff, '\0', sizeof(buff));

	/* fill the message header */
	hdr = (struct nlmsghdr *) buff;

	hdr->nlmsg_len = MESSAGE_SIZE;
	hdr->nlmsg_type = NLMSG_DONE;
	hdr->nlmsg_flags = 0;
	hdr->nlmsg_seq = 0;
	hdr->nlmsg_pid = getpid();

	/* the message */
	msg = (struct cn_msg *) NLMSG_DATA(hdr);
	msg->id.idx = CN_IDX_PROC;
	msg->id.val = CN_VAL_PROC;
	msg->seq = 0;
	msg->ack = 0;
	msg->len = sizeof(int);
	msg->data[0] = PROC_CN_MCAST_LISTEN;

	send(sock, hdr, hdr->nlmsg_len, 0);
}

/**
 * cn_fork_ignore - 
 */
static inline void cn_fork_ignore(int sock)
{
	char buff[128];		/* must be > MESSAGE_SIZE */
	struct nlmsghdr *hdr;
	struct cn_msg *msg;

	/* Clear the buffer */
	memset(buff, '\0', sizeof(buff));

	/* fill the message header */
	hdr = (struct nlmsghdr *) buff;

	hdr->nlmsg_len = MESSAGE_SIZE;
	hdr->nlmsg_type = NLMSG_DONE;
	hdr->nlmsg_flags = 0;
	hdr->nlmsg_seq = 0;
	hdr->nlmsg_pid = getpid();

	/* the message */
	msg = (struct cn_msg *) NLMSG_DATA(hdr);
	msg->id.idx = CN_IDX_PROC;
	msg->id.val = CN_VAL_PROC;
	msg->seq = 0;
	msg->ack = 0;
	msg->len = sizeof(int);
	msg->data[0] = PROC_CN_MCAST_IGNORE;

	send(sock, hdr, hdr->nlmsg_len, 0);
}

/**
 * invalid_pid - test if pid is a valid one
 * @pid: the pid to be tested
 *
 * Test if a process with ID #pid exists or not. We use the /proc/#pid
 * to test the validity. 
 * Function returns 1 if the pid is _NOT_ valid, 0 otherwise.
 */
static int invalid_pid(pid_t pid)
{
	char buffer[64];	/* should be enough ;) */
	DIR *dir;

	/* clean the buffer */
	memset(buffer, 0, sizeof(buffer));

	/* pid is valid if there is an entry in /proc */
	sprintf(buffer, "%s/%d", PROC_BASE, pid);
	if ((dir = opendir(buffer))) {
		closedir(dir);
		return 0;
	}

	return 1;		/* pid is not an existing process ID */
}

void free_list(gpointer data, gpointer user_data)
{
	g_list_free(data);
}

/**
 * cleanup - release memory used and unregister from module
 *
 * Release memory used by the hash table
 */
static void cleanup()
{
	close(sk_unix);
	close(sk_nl);
	unlink(JOBD_SK_PATH);

	/* release memory used by hash table and its items */
	g_hash_table_destroy(jobs_ht);
	/* release list that are linked in the jobs_list */
	g_list_foreach(jobs_list, &free_list, NULL);
	/* release the jobs_list */
	g_list_free(jobs_list);

	cn_fork_ignore(sk_nl);

	exit(EXIT_SUCCESS);
}

/**
 * update_jobs_list - Update the global variable jobs_list
 * @data: the job ID 
 * @user_data: a pointer to an ht_elmt
 */
void update_jobs_list(gpointer data, gpointer user_data)
{
	GList *list;
	unsigned int jid;
	pid_t pid;

	/* 
	 * the job exists, thus, there is already a list for this
	 * job in the jobs_list and it _MUST_ be at index = jid - 1 
	 * since index for GList starts from 0.
	 */
	jid = GPOINTER_TO_INT(data);
	list = g_list_nth_data(jobs_list, jid - 1);
	assert(list != NULL);

	pid = ((struct ht_elmt *)user_data)->pid;
	list = g_list_append(list, GINT_TO_POINTER(pid));

	JOBD_DEBUG("[update_jobs_list] pid %d in job %d", pid, jid);
}

/**
 * fork_handler - fork handler
 * @ppid: parent PID
 * @cpid: child PID
 *
 * It checks if the parent that creates a child is in a job. If it is, 
 * we add the child in the same job, otherwise we do nothing. 
 * It's the main property needed to do per group of processes 
 * accounting.
 */
static void fork_handler(int ppid, int cpid)
{
	struct ht_elmt *new_elmt;	/* the new element: the child */
	struct ht_elmt *parent;	/* the parent if there is one */

	/* 
	 * we check if the parent is already registered in the jobs 
	 * hash table 
	 */
	parent = (struct ht_elmt *)g_hash_table_lookup(jobs_ht, &ppid);

	if ((!parent) || (parent->pid != ppid))
		/* parent is not in a job so there is nothing to do */
		return;

	JOBD_DEBUG("[fork_handler] pid %d is in a job", ppid);

	/* create and add a new entry in the hash table */
	new_elmt = (struct ht_elmt *)malloc(sizeof(struct ht_elmt));
	if (!new_elmt) {
		syslog(LOG_INFO, "cannot allocate space for a job item");
		return;
	}

	/* Initialize values */
	new_elmt->pid = cpid;
	new_elmt->jobs = g_list_copy(parent->jobs);

	/* update the list of jobs to which process belongs */
	g_list_foreach(new_elmt->jobs, update_jobs_list, new_elmt);

	/* and finally, we update the hash table */
	g_hash_table_insert(jobs_ht, &new_elmt->pid, new_elmt);
}

/**
 * do_add_req - add a process in the jobs hash table
 * @pid: pid to be added in the hash table (used as key)
 * @jid: the job identifier
 *
 * Add a job item in the jobs hash table if it's not already in. If jid is 
 * equal to 0, it creates a new job.
 * Returns the jid or -1 if an error occured.
 */
static int do_add_req(pid_t pid, int jid)
{
	struct ht_elmt *elmtp;	/* a pointer to an element of the jobs hash table */

	JOBD_DEBUG("[do_add_req] BEGIN");

	if (invalid_pid(pid)) {
		syslog(LOG_INFO, "%d is not a valid pid", pid);
		return -1;
	}

	/* 
	 * to be valid, the job must exists. Otherwise, it must be created 
	 * automatically (jid == 0).
	 */
	if (jid >= next_jid) {
		syslog(LOG_INFO, "jid doesn't exist");
		return -1;
	}

	/* we check if pid is already registered in the jobs hash table */
	elmtp = (struct ht_elmt *)g_hash_table_lookup(jobs_ht, &pid);

	if (!elmtp || (elmtp->pid != pid)) {

		JOBD_DEBUG("process %d is not in the hash table", pid);

		/* create and add a new job item in the hash table */
		elmtp = (struct ht_elmt *)malloc(sizeof(struct ht_elmt));
		if (!elmtp) {
			syslog(LOG_INFO,
			       "cannot allocate space for a job item");
			return -1;
		}

		/* Initialize values */
		elmtp->pid = pid;
		elmtp->jobs = NULL;

		g_hash_table_insert(jobs_ht, &elmtp->pid, elmtp);
	}

	if (jid == 0) {
		int index;
		GList *new_list = NULL;

		/* get a new jid */
		jid = next_jid;
		next_jid++;

		/* 
		 * update the list of jobs to which the process 
		 * belongs to.
		 */
		elmtp->jobs = g_list_append(elmtp->jobs, GINT_TO_POINTER(jid));

		/* start a new list to be added to the global jobs_list */
		new_list = g_list_append(new_list, GINT_TO_POINTER(pid));
		jobs_list = g_list_append(jobs_list, new_list);

		/* index must be equal to next_jid - 1 (index starts at 0) */
		index = g_list_index(jobs_list, new_list);
		assert(index == (jid - 1));

		JOBD_DEBUG("[jid == 0] add #%d in job #%d", pid, jid);
	} else {
		/* check if pid is not already in the given job */
		if (!g_list_find(elmtp->jobs, GINT_TO_POINTER(jid))) {
			GList *list;

			JOBD_DEBUG("#%d is not in job #%d", pid, jid);

			elmtp->jobs =
			    g_list_append(elmtp->jobs, GINT_TO_POINTER(jid));

			/* the job already exists */
			list = g_list_nth(jobs_list, (jid - 1));
			/* 
			 * if list is NULL, it means that the job already 
			 * but it is empty. So we need to modify the
			 * data in the jobs_list. It sounds like it's not
			 * the good way to achive this and need to be fix
			 */
			list->data =
			    g_list_append(list->data, GINT_TO_POINTER(pid));
		}

		JOBD_DEBUG("[jid != 0] add #%d in job #%d", pid, jid);
	}

	JOBD_DEBUG("[do_add_req] END");
	return jid;
}

/**
 * do_remove_req - remove a process from all jobs
 * @pid: pid of the process to be removed
 * @jid: remove process from job #jid
 *
 * Remove the entry PID in the hash table. if jid is equal to 
 * 0, it removes it from all jobs otherwise, it removes it
 * from job #jid.
 */
static void do_remove_req(pid_t pid, int jid)
{
	struct ht_elmt *entry;	/* entry to remove */
	GList *list;

	JOBD_DEBUG("[do_remove_req] BEGIN");

	entry = (struct ht_elmt *)g_hash_table_lookup(jobs_ht, &pid);

	if (!entry || (entry->pid != pid)) {
		JOBD_DEBUG("[do_remove_req] entry not found");
		goto end;
	}

	/* we remove job #jid from the entry->jobs list */
	entry->jobs = g_list_remove(entry->jobs, GINT_TO_POINTER(jid));

	/* and we remove the process from job #jid */
	list = g_list_nth(jobs_list, jid - 1);
	if (!list) {
		JOBD_DEBUG("[do_remove_req] job is empty");
		goto end;
	}

	list->data = g_list_remove(list->data, GINT_TO_POINTER(pid));

	/* 
	 * if it doesn't belong to any jobs, we can remove it 
	 * from the hash table 
	 */
	if (g_list_length(entry->jobs) == 0) {
		g_list_free(entry->jobs);
		g_hash_table_remove(jobs_ht, &entry->pid);
		/* 
		 * don't need to free the entrey because the hash table was created
		 * using g_hash_table_new_full
		 */
	}

	JOBD_DEBUG("#%d removed from jid#%d", pid, jid);

      end:
	JOBD_DEBUG("[do_remove_req] END");
}

/**
 * __send_list_info - send info for a given pid
 * @data: a pid
 * @user_data:
 */
void __send_list_info(gpointer data, gpointer user_data)
{
	struct jobd_msg mesg;
	struct send_list_data *datap;
	pid_t pid;

	pid = GPOINTER_TO_INT(data);
	datap = (struct send_list_data *)user_data;

	mesg.req = GETINFO_REQ;
	mesg.pid = pid;
	mesg.jid = datap->index;

	JOBD_DEBUG("[__send_list_info] info about #%d in #%d", pid, mesg.jid);

	write(datap->sock, &mesg, sizeof(mesg));
}

/**
 * send_list_info - send info for a GList
 * @data: a list of pid
 * @user_data: 
 */
void send_list_info(gpointer data, gpointer user_data)
{
	struct send_list_data *datap = (struct send_list_data *)user_data;
	GList *list = data;

	/* data is NULL for empty jobs */
	if (list)
		g_list_foreach(list, &__send_list_info, user_data);

	datap->index++;
}

/**
 * do_getinfo_req - 
 * 
 * open a connection with a server using stream socket.
 */
static void do_getinfo_req(void)
{
	struct sockaddr_un client;
	int sock;
	int err;
	struct send_list_data data;

	sock = socket(AF_UNIX, SOCK_STREAM, 0);
	if (sock == -1) {
		syslog(LOG_INFO, "getinfo socket error");
		return;
	}

	client.sun_family = AF_UNIX;
	strcpy(client.sun_path, JOBD_EXCH_PATH);

	err =
	    connect(sock, (struct sockaddr *)&client,
		    sizeof(struct sockaddr_un));
	if (err == -1) {
		syslog(LOG_INFO, "getinfo connect error");
		close(sock);
		return;
	}

	/* Send information */
	data.index = 1;
	data.sock = sock;
	g_list_foreach(jobs_list, &send_list_info, &data);

	close(sock);
}

/**
 * closeall - close all FDs >= a specified value 
 * @fd: a file descriptor
 *
 * close all FDs wich are greater or equal to a system value
 */
static void closeall(int fd)
{
	int fdlimit = sysconf(_SC_OPEN_MAX);

	while (fd < fdlimit)
		close(fd++);
}

/**
 * destroy_data - destroy values in the global hash table
 * @data: data to be destroyed
 *
 * It is used by g_hash_table_destroy to free values that are dynamically 
 * allocated.
 */
void destroy_data(gpointer data)
{
	struct ht_elmt *elmtp = data;
	GList *list = elmtp->jobs;

	/* release memory used by the list of jobs */
	g_list_free(list);
	free(data);
}

/**
 * process_mesg - treat a message read from the message queue
 * @mesg: data to be treated
 * @toaddr: information about the receiver
 * 
 * Treat a message read from the message queue. The treatment depends on the 
 * command. 
 */
static void process_mesg(struct jobd_msg mesg, struct sockaddr_un toaddr)
{
	switch (mesg.req) {
	case ADD_REQ:
		do_add_req(mesg.pid, mesg.jid);
		break;
	case REMOVE_REQ:
		do_remove_req(mesg.pid, mesg.jid);
		break;
	case GETINFO_REQ:
		do_getinfo_req();
		break;
	case QUIT_REQ:
		cleanup();
		/* break; not needed because after cleanup it exits */
	default:
		syslog(LOG_INFO, "Unknown command");
	}

}

static void recv_sk_nl(int sk)
{
	char buff[1024];	/* it's large enough */
	struct nlmsghdr *hdr;
	struct cn_msg *msg;
	struct proc_event *ev;
	int len;
	int cpuid;
	int ppid;
	int ptgid;
	int cpid;
	int ctgid;

	/* Clear the buffer */
	memset(buff, '\0', sizeof(buff));

	/* Listen */
	len = recv(sk, buff, sizeof(buff), 0);
	if (len == -1) {
		syslog(LOG_INFO, "netlink recv error");
		return;
	}

	/* point to the message header */
	hdr = (struct nlmsghdr *)buff;

	switch (hdr->nlmsg_type) {
	case NLMSG_DONE:
		msg = (struct cn_msg *)NLMSG_DATA(hdr);
		ev = (struct proc_event *)msg->data;
		if (ev->what == PROC_EVENT_FORK) {
			cpuid = ev->cpu;
			ppid  = ev->event_data.fork.parent_pid;
			ptgid = ev->event_data.fork.parent_tgid;
			cpid  = ev->event_data.fork.child_pid;
			ctgid = ev->event_data.fork.child_tgid;
			JOBD_DEBUG("[seq#%d] %d %d - %d %d",
			    msg->seq, ppid, ptgid, cpid, ctgid);
			fork_handler(ppid, cpid);
		}
		break;
	case NLMSG_ERROR:
		syslog(LOG_INFO, "recv_sk_nl error");
		/* Fall through */
	default:
		break;

		return;
	}
}

static void recv_sk_unix(int sk)
{
	struct sockaddr_un fromaddr;
	struct jobd_msg mesg;

	recvfrom(sk, &mesg, sizeof(mesg), 0, (struct sockaddr *)&fromaddr,
		 NULL);

	JOBD_DEBUG("[req:%d pid:%d jid:%d]\n", mesg.req, mesg.pid, mesg.jid);
	process_mesg(mesg, fromaddr);

	return;
}

/**
 * daemonize - detach process from user and disappear into the background
 * @nochdir: set directory to "/" if not set
 * @noclose: close all opened file descriptors if not set
 *
 * returns -1 on failure, but you can't do much except exit in that case
 * since we may already have forked. This is based on the BSD version,
 * so the caller is responsible for things like the umask, etc.
 */
int daemonize(int nochdir, int noclose)
{
	/* 
	 * First, we fork to allow parent to exit. This returns control to 
	 * the command line. 
	 */
	switch (fork()) {
	case -1:
		/* error */
		return -1;
	case 0:
		/* child */
		break;
	default:
		/* Parent can exit */
		exit(EXIT_SUCCESS);
	}

	/* Now we're not a process group leader */
	if (setsid() < 0)	/* shoudn't fail */
		return -1;

	/* 
	 * New session is created and we are the leader. 
	 * Since a  controlling terminal is associated with a session, 
	 * and this new session has not yet acquired a controlling terminal 
	 * our process now has no controlling terminal, which is a Good Thing 
	 * for daemons.
	 * fork() again so the parent, (the session group leader), can exit.
	 * This means that we, as a non-session group leader, can never regain a
	 * controlling terminal.
	 */
	switch (fork()) {
	case -1:
		/* error */
		return -1;
	case 0:
		/* child */
		break;
	default:
		/* Parent can exit */
		exit(EXIT_SUCCESS);
	}

	umask(0);

	if (!nochdir)
		chdir("/");

	if (!noclose) {
		closeall(0);
		open("/dev/null", O_RDWR);
		dup(0);
		dup(0);
	}

	return 0;
}

/**
 * main - Principal routine and also a river rising in eastern Germany.
 *
 * Here are the different tasks:
 *    1) Become a daemon
 *    2) Open log file and run.
 *    3) Install signal handler
 *    4) Set next_id to 1
 *    5) Create the job hash table
 *    6) Create the message queue
 *    7) Run the infinite loop and wait message (or signal)
 */
int main(int argc, char **argv)
{
	int err;
	struct sockaddr_nl sa_nl;	/* info for the netlink interface */
	struct sockaddr_un sa_unix;	/* describs an PF_UNIX socket */
	fd_set fds;		/* file descriptor set */
	int max_fds;

	if (daemonize(0, 0) < 0) {
		perror("daemonize");
		exit(EXIT_FAILURE);
	}

	/* open connection to the system logger */
	openlog(log_ident, LOG_PID, LOG_INFO);

	/* Install signal handler */
	signal(SIGUSR1, cleanup);

	/* 
	 * set the first job identifier. As the jobs table is empty, the
	 * first job will be numbered 1.
	 */
	next_jid = 1;

	/* 
	 * Create the hash table where process will be store. Process ID 
	 * is the key of the hash table.
	 */
	jobs_ht =
	    g_hash_table_new_full(&g_int_hash, &g_int_equal, NULL,
				  &destroy_data);

	/* 
	 * Now we can start the communication.
	 * First, the jobd daemon listens for fork creation and it also 
	 * listens for user request. So we will use select() to be aware
	 * of the two events.
	 */

	/*
	 * The netlink communication 
	 * 
	 * Create an endpoint for communication. Use the kernel user
	 * interface device (PF_NETLINK) which is a datagram oriented
	 * service (SOCK_DGRAM). The netlink family used is
	 * NETLINK_CONNECTOR because we use the Process Events Connector
	 * patch.
	 */
	sk_nl = socket(PF_NETLINK, SOCK_DGRAM, NETLINK_CONNECTOR);
	if (sk_nl == -1) {
		syslog(LOG_INFO, "socket sk_nl error");
		return -1;
	}

	sa_nl.nl_family = AF_NETLINK;
	sa_nl.nl_groups = CN_IDX_PROC;
	sa_nl.nl_pid    = getpid();

	err = bind(sk_nl, (struct sockaddr *)&sa_nl, sizeof(sa_nl));
	if (err == -1) {
		syslog(LOG_INFO, "binding sk_nl error");
		close(sk_nl);
		return -1;
	}

	/*
	 * The local communication
	 *
	 * creates a socket for local interprocess communication 
	 */
	sk_unix = socket(AF_UNIX, SOCK_DGRAM, 0);
	if (sk_unix == -1) {
		syslog(LOG_INFO, "socket sk_unix error");
		close(sk_unix);
		close(sk_nl);
		return -1;
	}

	unlink(JOBD_SK_PATH);
	strcpy(sa_unix.sun_path, JOBD_SK_PATH);
	sa_unix.sun_family = AF_UNIX;

	err = bind(sk_unix, (struct sockaddr *)&sa_unix, sizeof(sa_unix));
	if (err == -1) {
		syslog(LOG_INFO, "binding sk_unix error");
		close(sk_unix);
		close(sk_nl);
		return -1;
	}

	/* We're ready to receive cn_fork netlink message */
	cn_fork_listen(sk_nl);
	
	/* Doing synchronous I/O multiplexing */
	max_fds = (sk_nl > sk_unix) ? sk_nl + 1 : sk_unix + 1;

	for (;;) {
		/* waiting for sk_nl and sk_unix socket changes */
		FD_ZERO(&fds);
		FD_SET(sk_nl, &fds);
		FD_SET(sk_unix, &fds);

		err = select(max_fds, &fds, NULL, NULL, NULL);

		if (err < 0) {
			syslog(LOG_INFO, " !!! select error");
			continue;
		}

		if (FD_ISSET(sk_unix, &fds))
			recv_sk_unix(sk_unix);

		if (FD_ISSET(sk_nl, &fds))
			recv_sk_nl(sk_nl);
	};

	return 0;
}
